/***************************************************************************
                          ctelnet.cpp  -  telnet...
    This file is a part of KMuddy distribution.
                             -------------------
    begin                : P Jun 14 2002
    copyright            : (C) 2002-2005 by Tomas Mecir
    email                : kmuddy@kmuddy.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#define CTELNET_CPP

#include "ctelnet.h"

#include "cactionmanager.h"
#include "cconnprefs.h"

#ifdef HAVE_SOUND
#include "cmsp.h"
#endif

// needed for removeSession which is called in disconnect
#include "kmuddy.h"

#ifdef HAVE_MXP
#include "cmxpmanager.h"
#endif

#include <klocale.h>
#include <kbufferedsocket.h>
#include <kmessagebox.h>
#include <kapplication.h>

extern KMuddy *kmuddy;

cTelnet::cTelnet (int sess) : cActionBase ("telnet", sess)
{
  socket = 0;
  termType = "KMuddy";
  datalen = 0;
  sentbytes = 0;
  offLineConnection=false;
  _connected = false;
  _connecting = false;
  curX = 0;
  curY = 0;
  _startupneg = true;
  setQuitCommand (QUIT);
  
  // initialize the socket
  createSocket ();

  MCCP = new cMCCP (this);

#ifdef HAVE_SOUND
  MSP = new cMSP (sess);
  usingmsp = false;
#endif

#ifdef HAVE_MXP
  usingmxp = false;
  mxpallow = 3;  //auto-detect
  mxpNegotiated = false;
#endif

  addEventHandler ("dimensions-changed", 20, PT_INT);
}

cTelnet::~cTelnet()
{
  if (isConnected ())
    disconnect ();

  removeEventHandler ("dimensions-changed");

  delete socket;
  delete MCCP;

#ifdef HAVE_SOUND
  delete MSP;
#endif
}

void cTelnet::eventHandler (QString event, int, int par1, int par2)
{
  if (event == "dimensions-changed") {
    windowSizeChanged (par1, par2);
  }
}

QString cTelnet::actionHandler (QString action, int,
    const QString &par1, const QString &)
{
  if (action == "set-quit-command")
    setQuitCommand (par1);
  return QString::null;
}

QString cTelnet::actionHandler (QString action, int, int par1, int)
{
  if (action == "mxp-allowed") {
#ifdef HAVE_MXP
    setMXPAllowed (par1);
#endif
  }
  else if (action == "set-negotiate-on-startup") {
    setNegotiateOnStartup (par1 != 0);
  }
  else if (action == "set-lp-mud-style") {
    setLPMudStyle (par1 != 0);
  }
  return QString::null;
}

void cTelnet::connectionFailed ()
{
#if KDE_IS_VERSION(3,4,0)
  if (socket->error() == KNetwork::KBufferedSocket::RemotelyDisconnected) {
    while (socket->bytesAvailable())
      socketRead ();
    return;
  }
#endif
  
  // don't crash on weird problems (happens on KDE 3.4, no idea why)
  if (socket->error() == KNetwork::KBufferedSocket::NoError)
    return;
 
  QString err = socket->errorString();
  cActionManager::self()->invokeEvent ("connection-failed", sess(), err);
  _connected = false;
  _connecting = false;
  // re-create socket - just in case ...
  delete socket;
  createSocket ();
}

/** establishes a new connection */
void cTelnet::connectIt (const QString &address, int port)
{
  // close existing connection first (if any)
  if (isConnected ())
    disconnect ();

  // handle offline connection
  if (isOffLineConnection())
  {
    _connected = true;
    _connecting = false;
    invokeEvent ("message", sess(), i18n ("--- A connection has been established ---"));
    invokeEvent ("connected", sess());
    return;
  }

  _connecting = true;
  cActionManager::self()->invokeEvent ("message", sess(), i18n ("Connecting ..."));
  socket->setTimeout (10000);
  hostName = address;
  hostPort = port;
  socket->connect (address, QString::number (port));
}

void cTelnet::socketFailed (int)
{
  connectionFailed ();
}

void cTelnet::createSocket ()
{
  socket = new KNetwork::KBufferedSocket;
  connect (socket, SIGNAL(connected(const KResolverEntry &)), this, SLOT (socketConnected ()));
  connect (socket, SIGNAL(gotError(int)), this, SLOT (socketFailed (int)));
  connect (socket, SIGNAL(readyRead()), this, SLOT (socketRead ()));
  connect (socket, SIGNAL(closed()), this, SLOT (socketClosed ()));
  socket->enableRead (true);
  socket->setInputBuffering (true);
  socket->setOutputBuffering (true);
}

void cTelnet::socketConnected ()
{
  if (_connected) return;

  _connected = true;
  _connecting = false;

  //reset MCCP, MSP and byte counters
  MCCP->reset ();
#ifdef HAVE_SOUND
  MSP->reset (hostName);
#endif

  //now we should be connected
  invokeEvent ("message", sess(), i18n ("--- A connection has been established ---"));
  cActionManager::self()->invokeEvent ("connected", sess());

#ifdef HAVE_MXP
  // MXP -must- be initialized AFTER the connected message, else it won't work
  setMXPAllowed (mxpallow);  //initialize MXP properly
#endif  
  sentbytes = 0;
  
  //prepare option variables
  for (int i = 0; i < 256; i++)
  {
    myOptionState[i] = false;
    hisOptionState[i] = false;
    announcedState[i] = false;
    heAnnouncedState[i] = false;
    triedToEnable[i] = false;
  }
  //reset buffer
  datalen = 0;
  
  //negotiate some telnet options, if allowed
  if (_startupneg)
  {
    QString s;
    
    //NAWS (used to send info about window size)
    s += (QChar) TN_IAC;
    s += (QChar) TN_WILL;
    s += (QChar) OPT_NAWS;
    //do not allow server to echo our text!
    s += (QChar) TN_IAC;
    s += (QChar) TN_DONT;
    s += (QChar) OPT_ECHO;
  
    doSendData (s, false);
  }
}

/** closes connection */
void cTelnet::disconnect ()
{
  if (!_connected) return;

  // wait a bit to see if we get any more data, process it if we do ...
  _connected = false;
  _connecting = false;

  datalen = 0;
  // send command quit first! - otherwise MUD would say you're net-dead...
  if (isConnected () && !isOffLineConnection())
    doSendData (quitcommand + "\n", true);

  if (isOffLineConnection())
    offLineConnection=false;

  //reset quit-command back to default value
  quitcommand = QUIT;

  socket->flush ();
  socket->closeNow ();
  delete socket;
  createSocket ();

  //alright - we're disconnected
  cActionManager::self()->invokeEvent ("disconnected", sess());

  // ------------------
  // TODO: the rest will not be here, but in their respective objects, hooked on the event
  
  //remove the session - should now be safe
  kmuddy->removeSession (sess(), true);
}

void cTelnet::setOffLineConnection (bool type)
{
  offLineConnection = type;
}

/** are we connected (offline connection is counted as connected) ? */
bool cTelnet::isConnected ()
{
  return (isOffLineConnection() || _connected);
}

#ifdef HAVE_SOUND

void cTelnet::setMSPGlobalPaths (const QStringList &paths)
{
  if (MSP != 0)
    MSP->setGlobalPaths (paths);
}

void cTelnet::setMSPAllowed (bool allow)
{
  if (MSP != 0)
    MSP->setMSPAllowed (allow);
}

void cTelnet::setDownloadAllowed (bool allow)
{
  if (MSP != 0)
    MSP->setDownloadAllowed (allow);
}

void cTelnet::processSoundRequest (bool isSOUND, QString fName, int volume, int repeats,
    int priority, QString type, QString url)
{
  if (MSP != 0)
    MSP->processRequest (isSOUND, fName, volume, repeats, priority, type, url);
}

#endif  //HAVE_SOUND

#ifdef HAVE_MXP

void cTelnet::setMXPAllowed (int allow)
{
  //update actual MXP usage...
  switch (allow) {
    case 1: //Never
    {
      callAction ("mxpmanager", "set-active", sess(), 0);
      //ask server to disable MXP
      QString s = (QChar) TN_IAC;
      s += (QChar) TN_DONT;
      s += (QChar) OPT_MXP;
      doSendData (s, false);
    }
    break;
    case 2: //If negotiated
    {
      callAction ("mxpmanager", "set-active", sess(), mxpNegotiated ? 1 : 0);
      callAction ("mxpmanager", "switch-open", sess());
    }
    break;
    case 3: //Auto-detect
    {
      callAction ("mxpmanager", "set-active", sess(), 1);
    }
    break;
    case 4: //Always
    {
      callAction ("mxpmanager", "set-active", sess(), 1);
      callAction ("mxpmanager", "switch-open", sess());
    }
    break;
  };
  mxpallow = allow;
}

#endif //HAVE_MXP

bool cTelnet::sendData (const QString &data)
{
  if (!(isConnected()))
    return false;
  // return true and dont send, since offline is used
  if (isOffLineConnection())
    return true;

  if (t_cmdEcho == true && t_lpmudstyle)
    prependGANewLine = false;

  //IAC byte must be doubled
  QString d;
  int len = data.length();
  bool gotIAC = false;
  for (int i = 0; i < len; i++)
    if ((unsigned char) data[i].latin1() == 255) {
      gotIAC = true;
      break;
    }
  if (gotIAC)  // double IACs, if any
    for (int i = 0; i < len; i++)
    {
      QChar ch = data[i];
      d += ch;
      if ((unsigned char) ch.latin1() == 255)
        d += ch;  //double IAC
    }

  //data ready, send it!
  // don't use 8-bit if we had to expand IACs
  return doSendData (gotIAC ? d : data, gotIAC ? false : true);
}

bool cTelnet::doSendData (const QString &data, bool use8Bit)
{
  if (!(isConnected()))
    return false;
  if (isOffLineConnection())
    return true;
  QString d = data;
  //write data to socket - it's so complicated because sometimes only a part of data
  //is accepted at a time
  int dataLength = data.length ();
  int remlen = dataLength;
  do {
    int written;
    if (use8Bit)
      written = socket->writeBlock (d.local8Bit (), remlen);
    else
      written = socket->writeBlock (d.latin1 (), remlen);
    if (written == -1)  // buffer full - try again
      continue;
    remlen -= written;
    d.remove (0, written);
  } while (remlen > 0);

  //update counter
  sentbytes += dataLength;
  return true;
}

void cTelnet::windowSizeChanged (int x, int y)
{
  //remember the size - we'll need it if NAWS is currently disabled but will
  //be enabled. Also remember it if no connection exists at the moment;
  //we won't be called again when connecting
  curX = x;
  curY = y;
  if (!(isConnected()))
    return;
  if (myOptionState[OPT_NAWS])   //only if we have negotiated this option
  {
    QString s;
    s = (QChar) TN_IAC;
    s += (QChar) TN_SB;
    s += (QChar) OPT_NAWS;
    unsigned char x1, x2, y1, y2;
    x1 = (unsigned char) x / 256;
    x2 = (unsigned char) x % 256;
    y1 = (unsigned char) y / 256;
    y2 = (unsigned char) y % 256;
    //IAC must be doubled
    s += (QChar) x1;
    if (x1 == TN_IAC)
      s += (QChar) TN_IAC;
    s += (QChar) x2; 
    if (x2 == TN_IAC)
      s += (QChar) TN_IAC;
    s += (QChar) y1;
    if (y1 == TN_IAC)
      s += (QChar) TN_IAC;
    s += (QChar) y2;
    if (y2 == TN_IAC)
      s += (QChar) TN_IAC;
    
    s += (QChar) TN_IAC;
    s += (QChar) TN_SE;
    doSendData (s, false);
  }
}

void cTelnet::processTelnetCommand (const QString &command)
{
  unsigned char ch = command[1];
    unsigned char option;
    switch (ch) {
    case TN_AYT:
      doSendData ("I'm here! Please be more patient!\r\n");
          //well, this should never be executed, as the response would probably
          //be treated as a command. But that's server's problem, not ours...
          //If the server wasn't capable of handling this, it wouldn't have
          //sent us the AYT command, would it? Impatient server = bad server.
          //Let it suffer! ;-)
      break;
    case TN_GA:
      recvdGA = true;
      //signal will be emitted later
      break;
    case TN_WILL:
      //server wants to enable some option (or he sends a timing-mark)...
      option = command[2];
      heAnnouncedState[option] = true;
      if (triedToEnable[option])
      {
        hisOptionState[option] = true;
        triedToEnable[option] = false;
      }
      else
      {
        if (!hisOptionState[option])
            //only if this is not set; if it's set, something's wrong wth the server
            //(according to telnet specification, option announcement may not be
            //unless explicitly requested)
        {
          if ((option == OPT_SUPPRESS_GA) || (option == OPT_STATUS) ||
              (option == OPT_TERMINAL_TYPE) || (option == OPT_NAWS))
                 //these options are supported; compression is handled
                 //separately
          {
            QString s;
            s = (QChar) TN_IAC;
            s += (QChar) TN_DO;
            s += (QChar) option;
            doSendData (s, false);
            hisOptionState[option] = true;
          }
#ifdef HAVE_MXP
          else
          if (option == OPT_MXP)
          {
            QString s;
            s = (QChar) TN_IAC;
            if (mxpallow >= 2)
              s += (QChar) TN_DO;
            else
              s += (QChar) TN_DONT;  //don't allow MXP if it's disabled
            s += (QChar) option;
            doSendData (s, false);
            hisOptionState[option] = true;

            //MXP is now negotiated
            mxpNegotiated = true;
            if (mxpallow == 2)  //MXP: if negotiated
            {
	      cMXPManager *mm = dynamic_cast<cMXPManager *>(object ("mxpmanager"));
              mm->setMXPActive (true);
            }
          }
#endif
#ifdef HAVE_SOUND
          else
          if (option == OPT_MSP)
          {
            QString s;
            s = (QChar) TN_IAC;
            s += (QChar) TN_DO;
            s += (QChar) option;
            doSendData (s, false);
            hisOptionState[option] = true;

            //MSP is now enabled
            usingmsp = true;
            MSP->enableMSP ();
          }
#endif  //HAVE_SOUND
          else
          if ((option == OPT_COMPRESS) || (option == OPT_COMPRESS2))
          //these are handled separately, as they're a bit special
          {
            if ((option == OPT_COMPRESS) && (hisOptionState[OPT_COMPRESS2]))
            {
              //protocol says: reject MCCP v1 if you have previously accepted
              //MCCP v2...
              QString s;
              s = (QChar) TN_IAC;
              s += (QChar) TN_DONT;
              s += (QChar) option;
              doSendData (s, false);
              hisOptionState[option] = false;
            }
            else
            {
              QString s;
              s = (QChar) TN_IAC;
              s += (QChar) TN_DO;
              s += (QChar) option;
              doSendData (s, false);
              hisOptionState[option] = true;
              //inform MCCP object about the change
              if ((option == OPT_COMPRESS))
                MCCP->setMCCP1 (true);
              else
                MCCP->setMCCP2 (true);
            }
          }
          else
          {
            QString s;
            s = (QChar) TN_IAC;
            s += (QChar) TN_DONT;
            s += (QChar) option;
            doSendData (s, false);
            hisOptionState[option] = false;
          }
        }
      }
      break;
    case TN_WONT:
      //server refuses to enable some option...
      option = command[2];
      if (triedToEnable[option])
      {
        hisOptionState[option] = false;
        triedToEnable[option] = false;
        heAnnouncedState[option] = true;
      }
      else
      {
        //send DONT if needed (see RFC 854 for details)
        if (hisOptionState[option] || (!heAnnouncedState[option]))
        {
          QString s = (QChar) TN_IAC;
          s += (QChar) TN_DONT;
          s += (QChar) option;
          doSendData (s, false);
          hisOptionState[option] = false;
          //inform MCCP object about the change - won't cause problems in
          //cMCCP - see cmccp.cpp for more info
          if ((option == OPT_COMPRESS))
            MCCP->setMCCP1 (false);
          if ((option == OPT_COMPRESS2))
            MCCP->setMCCP2 (false);
        }
        heAnnouncedState[option] = true;
      }
#ifdef HAVE_SOUND
      if (option == OPT_MSP)
      {
        //MSP is now disabled
        usingmsp = false;
        MSP->disableMSP ();
      }
#endif
#ifdef HAVE_MXP
      if (option == OPT_MXP)
      {
        //MXP is now disabled
        cMXPManager *mm = dynamic_cast<cMXPManager *>(object ("mxpmanager"));
        mm->setMXPActive (false);
      }
#endif
      break;
    case TN_DO:
      //server wants us to enable some option
      option = command[2];
      if (option == OPT_TIMING_MARK)
      {
        //send WILL TIMING_MARK
        QString s;
        s = (QChar) TN_IAC;
        s += (QChar) TN_WILL;
        s += (QChar) OPT_TIMING_MARK;
        doSendData (s, false);
      }
      else
      if (!myOptionState[256])
      //only if the option is currently disabled
      {
        if ((option == OPT_SUPPRESS_GA) || (option == OPT_STATUS) ||
            (option == OPT_TERMINAL_TYPE) || (option == OPT_NAWS))
        {

          QString s;
          s = (QChar) TN_IAC;
          s += (QChar) TN_WILL;
          s += (QChar) option;
          doSendData (s, false);
          myOptionState[option] = true;
          announcedState[option] = true;
        }
        else
        {
          QString s;
          s = (QChar) TN_IAC;
          s += TN_WONT;
          s += option;
          doSendData (s, false);
          myOptionState[option] = false;
          announcedState[option] = true;
        }
      }
      if (option == OPT_NAWS)  //NAWS here - window size info must be sent
        windowSizeChanged (curX, curY);
      break;
    case TN_DONT:
      //only respond if value changed or if this option has not been announced yet
      option = command[2];
      if (myOptionState[option] || (!announcedState[option]))
      {
        QString s;
        s = (QChar) TN_IAC;
        s += (QChar) TN_WONT;
        s += (QChar) option;
        doSendData (s, false);
        announcedState[option] = true;
      }
      myOptionState[option] = false;
      break;
    case TN_SB:
      //subcommand - we analyze and respond...
      option = command[2];
      switch (option) {
        case OPT_STATUS:
          //see OPT_TERMINAL_TYPE for explanation why I'm doing this
          if (true /*myOptionState[OPT_STATUS]*/)
          {
            if (command[3].latin1() == TNSB_SEND)
            //request to send all enabled commands; if server sends his
            //own list of commands, we just ignore it (well, he shouldn't
            //send anything, as we do not request anything, but there are
            //so many servers out there, that you can never be sure...)
            {
              QString s;
              s = (QChar) TN_IAC;
              s += (QChar) TN_SB;
              s += (QChar) OPT_STATUS;
              s += (QChar) TNSB_IS;
              for (int i = 0; i < 256; i++)
              {
                if (myOptionState[i])
                {
                  s += (QChar) TN_WILL;
                      //nice line, isn't it? :)
                  s += (QChar) ((unsigned char) i);
                }
                if (hisOptionState[i])
                {
                  s += (QChar) TN_DO;
                  s += (QChar) ((unsigned char) i);
                }
              }
              s += (QChar) TN_IAC;
              s += (QChar) TN_SE;
              doSendData (s, false);
            }
          }
        break;
        case OPT_TERMINAL_TYPE:
          if (myOptionState[OPT_TERMINAL_TYPE])
          {
            if (command[3].latin1() == TNSB_SEND)
              //server wants us to send terminal type; he can send his own type
              //too, but we just ignore it, as we have no use for it...
            {
              QString s;
              s = (QChar) TN_IAC;
              s += (QChar) TN_SB;
              s += (QChar) OPT_TERMINAL_TYPE;
              s += (QChar) TNSB_IS;
              s += termType;
              s += (QChar) TN_IAC;
              s += (QChar) TN_SE;
              doSendData (s, false);
            }
          }
        break;
        //other cmds should not arrive, as they were not negotiated.
        //if they do, they are merely ignored
      };
      break;
    //other commands are simply ignored (NOP and such, see .h file for list)
  };
}

void cTelnet::socketRead ()
{
  /*
  This function makes heavy use of char* instead of relying on QString,
  because otherwise there are problems with locale-based data and/or
  telnet commands.
  */

  // Looks like we often get this before getting the connected event ...
  if (_connecting && (!_connected))
    socketConnected();

  char buffer[32769];  //we should never receive such a big amount of data,
        //but we want to be sure that we read all we can...

  char buffer2[32769]; //decompression buffer

  int amount = socket->readBlock (buffer, 32768);
  if (amount == -1)
    return;   //something is wrong (no data?)
  if (amount == 0)
    return;   //0 means socket has been closed; maybe I'll do something more?
  buffer[amount] = '\0';    //just to be sure - mark end of string

  invokeEvent ("raw-data-comp", sess(), QString::fromLocal8Bit (buffer));

  //we'll need cConnPrefs later on
  cConnPrefs *cp = dynamic_cast<cConnPrefs *>(object ("connprefs"));

  //MCCP will be handled first. This is done in a separate class, which
  //detects all MCCP-related codes itself. I know that this involves some
  //code duplication (two telnet option parsers), but it's needed because
  //telnet sequences may (and will) appear inside the compressed streams,
  //so it's best to get plain uncompressed data first. Another reason is
  //improper design of MCCP v1, which uses unterminated telnet subsequences.
  //This has been fixed in MCCP v2, but some MUDs may still use the old
  //version.

  MCCP->prepareDecompression (buffer, buffer2, amount, 32768);
  int size;
  while ((size = MCCP->uncompressNext ()) != -1)
  {
    //inform plug-ins about uncompressed data (can be equal to raw data if MCCP isn't used
    invokeEvent ("raw-data", sess(), QString::fromLocal8Bit (buffer2));

    //we add buffer to existing data (if any)... I do this, because previous
    //data could end in the middle of a telnet command, which was then left
    //in the buffer for subsequent reading (= this one ;)
    for (int i = 0; i < size; i++)
      data[datalen + i] = buffer2[i];
    datalen += size;
    data[datalen] = 0;

    //we have some data...
    newdata = true;

    //clear the GO-AHEAD flag
    recvdGA = false;
    
    //now we have the data, but we cannot forward it to next stage of processing,
    //because the data contains telnet commands
    //so we parse the text and process all telnet commands:

    QString cleandata, command;
    bool iac, iac2, insb;
      //iac: last char was IAC
      //iac2: last char was DO, DONT, WILL or WONT
      //insb: we're in IAC SB, waiting for IAC SE
    int processed = -1;
    iac = iac2 = insb = false;
    cleandata = "";
    command = "";

    for (unsigned int i = 0; i < datalen; i++)
    {
      unsigned char ch = data[i];
      if (iac || iac2 || insb || (ch == TN_IAC))
      {
        //there are many possibilities here:
        //1. this is IAC, previous character was regular data
        if (! (iac || iac2 || insb) && (ch == TN_IAC))
        {
          iac = true;
          command += ch;
        }
        else
        //2. seq. of two IACs
        if (iac && (ch == TN_IAC) && (!insb))
        {
          iac = false;
          char str[2];
          str[0] = ch;
          str[1] = 0;
          cleandata += QString::fromLocal8Bit (str);
          processed = i;
          command = "";
        }
        else
        //3. IAC DO/DONT/WILL/WONT
        if (iac && (!insb) &&
            ((ch == TN_WILL) || (ch == TN_WONT) || (ch == TN_DO) || (ch == TN_DONT)))
        {
          iac = false;
          iac2 = true;
          command += ch;
        }
        else
        //4. IAC DO/DONT/WILL/WONT <command code>
        if (iac2)
        {
          iac2 = false;
          command += ch;
          processTelnetCommand (command);
          processed = i;
          command = "";
        }
        else
        //5. IAC SB
        if (iac && (!insb) && (ch == TN_SB))
        {
          iac = false;
          insb = true;
          command += ch;
        }
        else
        //6. IAC SE without IAC SB - error - ignored
        if (iac && (!insb) && (ch == TN_SE))
        {
          command = "";
          iac = false;
        }
        else
        //7. inside IAC SB
        if (insb)
        {
          command += ch;
          if (iac && (ch == TN_SE))  //IAC SE - end of subcommand
          {
            processTelnetCommand (command);
            processed = i;
            command = "";
            iac = false;
            insb = false;
          }
          if (iac)
            iac = false;
          else
          if (ch == TN_IAC)
            iac = true;
        }
        else
        //8. IAC fol. by something else than IAC, SB, SE, DO, DONT, WILL, WONT
        {
          iac = false;
          command += ch;
          processTelnetCommand (command);
          //this could have set receivedGA to true; we'll handle that later
          // (at the end of this function)
          processed = i;
          command = "";
        }
      }
      else   //plaintext
      {
        //everything except CRLF is okay; CRLF is replaced by LF(\n) (CR ignored)
        if (ch != 13)
        {
          char str[2];
          str[0] = ch;
          str[1] = 0;
          cleandata += QString::fromLocal8Bit (str);
          processed = i;
        }
      }

      //we've just received the GA signal - higher layers shall be informed about it
      if (recvdGA)
      {
        //ask MSP parser to look for MSP tags if MSP is enabled
#ifdef HAVE_SOUND
        //look if the user want to have MSP all the time
        bool alwaysmsp = false;
        if (cp)
          alwaysmsp = cp->alwaysMSP();
        //hand data to MSP parser if desired
        if (usingmsp || alwaysmsp)
          cleandata = MSP->parseServerOutput (cleandata);
#endif

        //prepend a newline, if needed
        if (prependGANewLine && t_lpmudstyle)
          cleandata = "\n" + cleandata;
        prependGANewLine = false;
        //forward data for further processing
        invokeEvent ("data-received", sess(), cleandata);

        if (cp && cp->consolePrompt())
          //we'll need to prepend a new-line in next data sending
          prependGANewLine = true;
        //we got a prompt
        invokeEvent ("received-ga", sess());

        //clean the flag, and the data (so that we don't send it multiple times)
        cleandata = QString::null;
        recvdGA = false;
      }
    }

    //shift data (so that telnet commands broken into multiple packets
    // are handled correctly)
    for (unsigned int i = 0; i < datalen - (processed + 1); i++)
      data[i] = data[(processed + 1) + i];
    datalen -= (processed + 1);

    //some data left to send - do it now!
    if (!cleandata.isEmpty())
    {
      //ask MSP parser to look for MSP tags if MSP is enabled
#ifdef HAVE_SOUND
      //look if the user want to have MSP all the time
      bool alwaysmsp = false;
      if (cp)
        alwaysmsp = cp->alwaysMSP();
      //hand data to MSP parser if desired
      if (usingmsp || alwaysmsp)
        cleandata = MSP->parseServerOutput (cleandata);
#endif

      //prepend a newline, if needed
      if (prependGANewLine && t_lpmudstyle)
        cleandata = "\n" + cleandata;
      prependGANewLine = false;
      //forward data for further processing
      invokeEvent ("data-received", sess(), cleandata);
    }
  }

  invokeEvent ("text-here", sess());
}

void cTelnet::socketClosed ()
{
  disconnect ();
  (KApplication::kApplication())->processEvents ();
}

void cTelnet::setQuitCommand (QString cmd)
{
  quitcommand = cmd;
}

void cTelnet::setCommandEcho (bool cmdEcho)
{
  t_cmdEcho = cmdEcho;
}

void cTelnet::setLPMudStyle (bool lpmudstyle)
{
  t_lpmudstyle = lpmudstyle;
}

void cTelnet::setNegotiateOnStartup (bool startupneg)
{
  _startupneg = startupneg;
}

